from redis import Redis
import os


class RedisHrefPool(object):
    def __init__(self, name: str, host: str, pwd: str, port=6379, db=0):
        self.name = name
        self.address = '%s:%d' % (host, port)
        self._pool = Redis(host=host, port=port, db=db, password=pwd)

    @property
    def all_href(self):
        return self._get_all_href()

    def _get_all_href(self):
        return self._pool.smembers(self.name)

    def insert(self, href: str):
        if not self.is_duplicate(href):
            self._pool.sadd(self.name, href)

    def multi_insert(self, href_list: list):
        try:
            for href in href_list:
                self.insert(href)
        except Exception as e:
            print(e)
            return

    def is_duplicate(self, href) -> bool:
        return self._pool.sismember(self.name, href)

    def serialize(self):
        pass


class MemPool(object):
    def __init__(self, name: str):
        self.name = name
        self._pool = list()
        if '%s_pool' % name in os.listdir(os.path.abspath(os.curdir)):
            with open('./%s_pool' % name, 'r')as f:
                url_list = f.readlines()
                for url in url_list:
                    url = url.replace('\n', '')
                    self.insert(url)

    @property
    def all_href(self):
        return self._get_all_href()

    def _get_all_href(self):
        return self._pool

    def insert(self, href: str):
        if not self.is_duplicate(href):
            self._pool.append(href)

    def multi_insert(self, href_list: list):
        try:
            for href in href_list:
                self.insert(href)
        except Exception as e:
            print(e)
            return

    def is_duplicate(self, href) -> bool:
        return href in self._pool

    def serialize(self):
        with open('./%s_pool' % self.name, 'w')as f:
            for url in self.all_href:
                f.write(url)
                f.write('\n')


from hannibal.parser import IncreasingParser, BaseParser
from hannibal.spider import DistributeCollector, LocalCollector
from hannibal.util import RedisQueue, RedisHrefPool, MemPool, MemQueue, TrickHelper, extract_json, extract_html, Mission



from multiprocessing import Queue
from queue import Empty
from redis import Redis
from hannibal.util.common_utils import Mission
import uuid
import os


class RedisQueue(object):
    def __init__(self, name: str, host: str, pwd: str, port=6379, db=0, limited=False):
        self.name = name
        self.address = '%s:%d' % (host, port)
        self._queue = Redis(host=host, port=port, db=db, password=pwd)
        self.limited = limited
        self.endpoint = None

    @property
    def queue_size(self):
        return self._get_queue_size()

    def _get_queue_size(self):
        return self._queue.llen(self.name)

    def enqueue(self, value: str):
        self._queue.rpush(self.name, value)

    def dequeue(self):
        element = self._queue.lpop(self.name)
        return element if element else None

    def init_queue(self, mission_list: list):
        for mission in mission_list:
            self.enqueue(mission.serialize())
        if self.limited:
            self.endpoint = str(uuid.uuid4())
            self.enqueue(self.endpoint)

    def serialize(self):
        pass


class MemQueue(object):
    _CACHE_LIST = []

    def __init__(self, name: str, limited=False):
        self.name = name
        self._queue = Queue()
        self.limited = limited
        self.endpoint = None
        if '%s_queue' % name in os.listdir(os.path.abspath(os.curdir)):
            with open('./%s_queue' % name, 'r')as f:
                mission_list = f.readlines()
                for mission_str in mission_list:
                    mission_str = mission_str.replace('\n', '')
                    self.enqueue(mission_str)
                    MemQueue._CACHE_LIST.append(Mission.deserialize(mission_str))

    @property
    def queue_size(self):
        return self._get_queue_size()

    def _get_queue_size(self):
        return self._queue.qsize()

    def enqueue(self, value: str):
        self._queue.put(value)

    def dequeue(self, timeout=None):
        element = self._queue.get(timeout=timeout)
        return element if element else None

    def init_queue(self, mission_list: list):
        unique_tag_list = [m.unique_tag for m in MemQueue._CACHE_LIST]
        if '%s_queue' % self.name not in os.listdir(os.path.abspath(os.curdir)):
            for mission in mission_list:
                if mission.unique_tag not in unique_tag_list:
                    self.enqueue(mission.serialize())
            if self.limited:
                self.endpoint = str(uuid.uuid4())
                self.enqueue(self.endpoint)

    def serialize(self):
        with open('./%s_queue' % self.name, 'w')as f:
            while 1:
                try:
                    url = self.dequeue(timeout=1)
                    if self.endpoint:
                        if url == self.endpoint:
                            break
                except Empty:
                    break
                else:
                    f.write(str(url))
                    f.write('\n')


import time
import asyncio
from threading import Thread

try:
    import uvloop

    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
except ImportError:
    pass


class BaseParser(object):
    _BEFORE_PARSE_MIDDLEWARE_LIST = []
    _AFTER_PARSE_MIDDLEWARE_LIST = []

    def __init__(self, mission_queue, parse_queue, href_pool):
        self.mission_queue = mission_queue
        self.parse_queue = parse_queue
        self.href_pool = href_pool
        self.meta_url = None
        self.url_pattern = None
        self.parse_loop = asyncio.new_event_loop()

    def pre_process(self, response_object):
        """Method for processing the response content into content object,
        which could be achieved by RegEx, BeautifulSoup or PyQuery."""
        return response_object

    async def parse_page(self, response_content):
        content_object = self.pre_process(response_content)
        extract_data_task = asyncio.ensure_future(self.extract_data(content_object), loop=self.parse_loop)
        asyncio.run_coroutine_threadsafe(await extract_data_task, loop=self.parse_loop)

    async def extract_data(self, content_object) -> None:
        """extract data from content"""
        raise NotImplementedError

    @staticmethod
    def _start_loop(loop):
        asyncio.set_event_loop(loop)
        loop.run_forever()

    def conquer(self):
        t = Thread(target=self._start_loop, args=(self.parse_loop,))
        t.setDaemon(True)
        t.start()
        try:
            while 1:
                if self.parse_queue.queue_size == 0:
                    time.sleep(10)
                else:
                    html_content = self.parse_queue.dequeue()
                    if html_content:
                        asyncio.run_coroutine_threadsafe(self.parse_page(html_content), loop=self.parse_loop)
        except KeyboardInterrupt:
            self.mission_queue.serialize()
            self.href_pool.serialize()
        finally:
            pass


class IncreasingParser(object):
    def __init__(self, mission_queue, parse_queue, href_pool):
        self.mission_queue = mission_queue
        self.parse_queue = parse_queue
        self.href_pool = href_pool
        self.url_pattern = None
        self.parse_loop = asyncio.new_event_loop()

    def pre_process(self, response_object):
        """Method for processing the response content into content object,
        which could be achieved by RegEx, BeautifulSoup or PyQuery."""
        return response_object

    async def parse_page(self, response_content):
        content_object = self.pre_process(response_content)

        extract_data_task = asyncio.ensure_future(self.extract_data(content_object), loop=self.parse_loop)
        asyncio.run_coroutine_threadsafe(await extract_data_task, loop=self.parse_loop)

        process_task = asyncio.ensure_future(self.extract_sub_href(content_object), loop=self.parse_loop)
        asyncio.run_coroutine_threadsafe(await process_task, loop=self.parse_loop)

    def go_forward(self, content_object) -> bool:
        """return whether go forward"""
        raise NotImplementedError

    def extract_sub_list(self, content_object) -> list:
        """return sub href list"""
        raise NotImplementedError

    async def extract_sub_href(self, content_object):
        sub_mission_list = self.extract_sub_list(content_object)
        if self.go_forward(content_object):
            for mission in sub_mission_list:
                if not self.href_pool.is_duplicate(mission.unique_tag):
                    self.mission_queue.enqueue(mission.serialize())

    async def extract_data(self, content_object):
        """extract data from content"""
        raise NotImplementedError

    @staticmethod
    def _start_loop(loop):
        asyncio.set_event_loop(loop)
        loop.run_forever()

    def conquer(self):
        t = Thread(target=self._start_loop, args=(self.parse_loop,))
        t.setDaemon(True)
        t.start()
        try:
            while 1:
                if self.parse_queue.queue_size == 0:
                    time.sleep(10)
                else:
                    html_content = self.parse_queue.dequeue()
                    if html_content:
                        asyncio.run_coroutine_threadsafe(self.parse_page(html_content), loop=self.parse_loop)
        except KeyboardInterrupt:
            self.mission_queue.serialize()
            self.href_pool.serialize()
        finally:
            pass



from hannibal.util.href_pool import RedisHrefPool, MemPool
from hannibal.util.queues import RedisQueue, MemQueue
from hannibal.util.tricks import TrickHelper
from hannibal.util.common_utils import extract_json, extract_html, Mission



import random
from queue import Queue
from threading import Thread


class TrickHelper(object):
    HEADERS = [
        'Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_6_8; en-us) AppleWebKit/534.50 (KHTML, like Gecko) Version/5.1 Safari/534.50',
        'Mozilla/5.0 (Windows NT 10.0; WOW64; rv:38.0) Gecko/20100101 Firefox/38.0',
        'IE 9.0User-Agent:Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; Trident/5.0;',
        'Opera/9.80 (Windows NT 6.1; U; en) Presto/2.8.131 Version/11.11',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_7_0) AppleWebKit/535.11 (KHTML, like Gecko) Chrome/17.0.963.56 Safari/535.11',
        'Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 5.1; TencentTraveler 4.0)',
        'Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 5.1; Trident/4.0; SE 2.X MetaSr 1.0; SE 2.X MetaSr 1.0; .NET CLR 2.0.50727; SE 2.X MetaSr 1.0)',
        'Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 5.1; 360SE)',
        'Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 5.1)',
        'Mozilla/5.0 (Windows NT 10.0; WOW64; Trident/7.0; .NET4.0C; .NET4.0E; .NET CLR 2.0.50727; .NET CLR 3.0.30729; .NET CLR 3.5.30729; InfoPath.3; rv:11.0) like Gecko',
        'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/536.11 (KHTML, like Gecko) Chrome/20.0.1132.11 TaoBrowser/2.0 Safari/536.11',
        'Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; WOW64; Trident/5.0; SLCC2; .NET CLR 2.0.50727; .NET CLR 3.5.30729; .NET CLR 3.0.30729; Media Center PC 6.0; .NET4.0C; .NET4.0E; LBBROWSER) ',
        'Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; WOW64; Trident/5.0; SLCC2; .NET CLR 2.0.50727; .NET CLR 3.5.30729; .NET CLR 3.0.30729; Media Center PC 6.0; .NET4.0C; .NET4.0E; QQBrowser/7.0.3698.400)',
        'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Maxthon/4.4.3.4000 Chrome/30.0.1599.101 Safari/537.36'
    ]

    def __init__(self, pool_size=100, use_https=False, use_proxy=False, proxy_provider=None, referer=None):
        self.use_proxy = use_proxy
        self._proxy_pool = None
        self.use_https = use_https
        self.proxy_prefix = 'https://' if self.use_https else 'http://'
        self.lock = False
        self.pool_size = pool_size
        self.threshold = int(pool_size / 2)
        self.proxy_provider = proxy_provider
        if self.use_proxy and not proxy_provider:
            raise AttributeError(
                "Proxy provider must be configured if you would like to use proxy server for request redirection.")
        elif self.use_proxy:
            self._proxy_pool = Queue()
            self.register_proxy_provider(proxy_provider)
            self.update_pool()
        self.referer = referer

    def register_proxy_provider(self, proxy_provider):
        if not hasattr(proxy_provider, '__call__'):
            raise AttributeError('Given Provider is not callable')
        else:
            self.proxy_provider = proxy_provider

    def trick(self):
        agent = self.get_random_agent()
        trick_dict = {'User-Agent': agent}
        if self.referer:
            trick_dict['Referer'] = self.referer
        return trick_dict

    @staticmethod
    def get_random_agent():
        return TrickHelper.HEADERS[random.randint(0, len(TrickHelper.HEADERS) - 1)]

    def get_proxy(self):
        if not self.use_proxy:
            return None
        else:
            try:
                proxy_ip = self._proxy_pool.get()
            except Exception as e:
                print(e)
                return None
            else:
                http_request_proxy = self.proxy_prefix + proxy_ip
                if self._proxy_pool.qsize() <= self.threshold:
                    Thread(target=self.update_pool()).start()
                return http_request_proxy

    def update_pool(self):
        if not self.proxy_provider:
            raise NotImplementedError("Proxy provider function needed to be implemented.")
        elif not self.lock:
            self.lock = True
            proxy_list = self.proxy_provider()
            for proxy in proxy_list:
                self._proxy_pool.put(proxy)


if __name__ == '__main__':
    t = TrickHelper()
    print(t.trick())


import aiohttp
import asyncio
from queue import Queue
from collections import deque
from threading import Thread
from hannibal.util import TrickHelper, Mission

try:
    import uvloop

    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
except ImportError:
    pass


class LocalCollector(object):
    _BEFORE_COLLECT_MIDDLEWARE_LIST = deque()
    _ERROR_HANDLER_DICT = dict()

    def __init__(self, mission_queue, parse_function, href_pool, seed_mission=None, cache_size=3, *args, **kwargs):
        self.mission_queue = mission_queue
        self.parse_function = parse_function
        self.href_pool = href_pool
        if seed_mission:
            self.mission_queue.enqueue(seed_mission.serialize())
        self.trick_helper = TrickHelper()
        self.collect_loop = asyncio.new_event_loop()
        self.semaphore = asyncio.Semaphore(cache_size, loop=self.collect_loop)
        self.client_session = aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(verify_ssl=False, loop=self.collect_loop), loop=self.collect_loop)
        self.collect_queue = Queue(maxsize=cache_size)

    @staticmethod
    def register_error_handler(code, handler, *args, **kwargs):
        assert isinstance(code, int)
        assert callable(handler)
        LocalCollector._ERROR_HANDLER_DICT[code] = handler

    @staticmethod
    def register_middleware(middleware, *args, **kwargs):
        assert callable(middleware)
        LocalCollector._BEFORE_COLLECT_MIDDLEWARE_LIST.append(middleware)

    async def collect(self, *args, **kwargs):
        with (await self.semaphore):
            mission_str = self.collect_queue.get()
            if isinstance(mission_str, bytes):
                mission_str = mission_str.decode('utf-8')
            mission = Mission.deserialize(mission_str)
            if not mission:
                raise AttributeError("Mission string parsing failed.")
            else:
                for func in LocalCollector._BEFORE_COLLECT_MIDDLEWARE_LIST:
                    func(mission)
                if not self.href_pool.is_duplicate(mission.unique_tag):
                    await self._collect(mission)

    def _insert_url(self, future):
        processed_url = future.result()
        if processed_url:
            self.href_pool.insert(processed_url)

    async def _collect(self, mission):
        url = mission.url
        headers = self.trick_helper.trick()
        async with self.client_session.request(mission.method, url,
                                               **{mission.data_type: mission.data, 'headers': headers})as response:
            process_task = asyncio.ensure_future(self.process_response(response, mission), loop=self.collect_loop)
            process_task.add_done_callback(self._insert_url)
            asyncio.run_coroutine_threadsafe(await process_task, loop=self.collect_loop)

    async def process_response(self, response, mission):
        status_code = int(response.status)
        if 400 < status_code < 600:
            handler = LocalCollector._ERROR_HANDLER_DICT.get(status_code, None)
            if handler:
                handler(response, mission)
            return None
        else:
            await self.parse_function(response)
            return mission.unique_tag

    def _pop_url(self) -> str:
        try:
            url = self.mission_queue.dequeue()
            return url if url else ''
        except Exception as e:
            print(e)
            return ''

    @staticmethod
    def _start_loop(loop):
        asyncio.set_event_loop(loop)
        loop.run_forever()

    def conquer(self, limited=False):
        t = Thread(target=self._start_loop, args=(self.collect_loop,))
        t.setDaemon(True)
        t.start()

        try:
            while 1:
                if not self.collect_queue.full():
                    url = self._pop_url()
                    if url:
                        self.collect_queue.put(url)
                    if limited and url == self.mission_queue.endpoint:
                        break
                    if url:
                        asyncio.run_coroutine_threadsafe(self.collect(), loop=self.collect_loop)
        except KeyboardInterrupt:
            self.mission_queue.serialize()
            self.href_pool.serialize()
        finally:
            asyncio.run_coroutine_threadsafe(self.client_session.close(), loop=self.collect_loop)



import json


class Mission(object):
    def __init__(self, unique_tag, url, method='GET', data=None, data_type='data'):
        self.unique_tag = str(unique_tag)
        self.url = url
        if method not in ['GET', 'POST', 'PUT', 'DELETE']:
            raise ValueError('invalid method')
        else:
            self.method = method
        self.data = data
        if data_type not in ['json', 'data']:
            raise ValueError('invalid data type')
        else:
            self.data_type = data_type

    def serialize(self):
        return json.dumps(self.__dict__)

    @classmethod
    def deserialize(cls, str_obj):
        try:
            assert isinstance(str_obj, str)
            mission_dict = json.loads(str_obj)
            assert isinstance(mission_dict, dict)
        except AssertionError:
            return None
        else:
            mission = cls(
                unique_tag=mission_dict.get('unique_tag'),
                url=mission_dict.get('url'),
                method=mission_dict.get('method'),
                data=mission_dict.get('data'),
                data_type=mission_dict.get('data_type')
            )
            return mission


async def extract_html(response, encoding='utf-8'):
    return await response.text(encoding=encoding)


async def extract_json(response, encoding='utf-8'):
    return await response.json(encoding=encoding)



from setuptools import setup, find_packages

setup(
    name='hannibal',
    version='0.0.7',
    author='JorgenLiu',
    author_email='avalon852456@gmail.com',
    url='https://github.com/JorgenLiu/hannibal',
    maintainer='JorgenLiu',
    maintainer_email='avalon852456@gmail.com',
    license='MIT License',
    description=u'A light-weight crawler framework based on asyncio, aiohttp and redis.',
    packages=find_packages(),
    install_requires=[
        'aiohttp',
        'asyncio',
        'uvloop',
        'redis',
        'requests',
        'async-timeout'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries'
    ]
)
